import logging
from pathlib import Path
from typing import Literal, Sequence, Union

import torch
from iohub import open_ome_zarr
from lightning.pytorch import LightningDataModule, LightningModule, Trainer
from lightning.pytorch.utilities.compile import _maybe_unwrap_optimized
from torch.onnx import OperatorExportTypes

from viscy.utils.meta_utils import generate_normalization_metadata


class VSTrainer(Trainer):
    def preprocess(
        self,
        data_path: Path,
        channel_names: Union[list[str], Literal[-1]] = -1,
        num_workers: int = 1,
        block_size: int = 32,
        model: LightningModule = None,
        datamodule: LightningDataModule = None,
        dataloaders: Sequence = None,
    ):
        """Compute dataset statistics before training or testing for normalization.

        :param Path data_path: Path to the HCS OME-Zarr dataset
        :param Union[list[str], Literal[ channel_names: channel names,
            defaults to -1 (all channels)
        :param int num_workers: number of workers, defaults to 1
        :param int block_size: sampling block size, defaults to 32
        :param LightningModule model: place holder for model, ignored
        :param LightningDataModule datamodule: place holder for datamodule, ignored
        :param Sequence dataloaders: place holder for dataloaders, ignored
        """
        if model or dataloaders or datamodule:
            logging.debug("Ignoring model and data configs during preprocessing.")
        with open_ome_zarr(data_path, layout="hcs", mode="r") as dataset:
            channel_indices = (
                [dataset.channel_names.index(c) for c in channel_names]
                if channel_names != -1
                else channel_names
            )
        generate_normalization_metadata(
            zarr_dir=data_path,
            num_workers=num_workers,
            channel_ids=channel_indices,
            grid_spacing=block_size,
        )

    def export(
        self,
        model: LightningModule,
        export_path: str,
        ckpt_path: str,
        format="onnx",
        datamodule: LightningDataModule = None,
        dataloaders: Sequence = None,
    ):
        """Export the model for deployment (currently only ONNX is supported).

        :param LightningModule model: module to export
        :param str export_path: output file name
        :param str ckpt_path: model checkpoint
        :param str format: format (currently only ONNX is supported), defaults to "onnx"
        :param LightningDataModule datamodule: placeholder for datamodule,
            defaults to None
        :param Sequence dataloaders: placeholder for dataloaders, defaults to None
        """
        if dataloaders or datamodule:
            logging.debug("Ignoring datamodule and dataloaders during export.")
        if not format.lower() == "onnx":
            raise NotImplementedError(f"Export format '{format}'")
        model = _maybe_unwrap_optimized(model)
        self.strategy._lightning_module = model
        model.load_state_dict(torch.load(ckpt_path)["state_dict"])
        model.eval()
        model.to_onnx(
            export_path,
            input_sample=model.example_input_array,
            export_params=True,
            opset_version=18,
            operator_export_type=OperatorExportTypes.ONNX_ATEN_FALLBACK,
            input_names=["input"],
            output_names=["output"],
            dynamic_axes={
                "input": {
                    0: "batch_size",
                    1: "channels",
                    3: "num_rows",
                    4: "num_cols",
                },
                "output": {
                    0: "batch_size",
                    1: "channels",
                    3: "num_rows",
                    4: "num_cols",
                },
            },
        )
        logging.info(f"ONNX exported at {export_path}")



# script to profile dataloading

from profilehooks import profile

from viscy.data.hcs import HCSDataModule

dataset = "/path/to/dataset.zarr"


dm = HCSDataModule(
    dataset,
    "Phase3D",
    "Deconvolved-Nuc",
    5,
    0.8,
    batch_size=32,
    num_workers=32,
    augment=None,
    caching=False,
)

dm.setup("fit")


@profile(immediate=True, sort="time", dirs=True)
def load_batch(n=1):
    for i, batch in enumerate(dm.train_dataloader()):
        print(batch["source"].shape)
        print(dm.on_before_batch_transfer(batch, 0)["target"].shape)
        if i == n - 1:
            break


load_batch(3)



from pathlib import Path

import numpy as np
from iohub import open_ome_zarr
from pytest import TempPathFactory, fixture

channel_names = ["Phase", "Retardance", "GFP", "DAPI"]


def _build_hcs(path: Path, channel_names: list[str], zyx_shape: tuple[int, int, int]):
    dataset = open_ome_zarr(
        path,
        layout="hcs",
        mode="w",
        channel_names=channel_names,
    )
    for row in ("A", "B"):
        for col in ("1", "2"):
            for fov in ("0", "1"):
                pos = dataset.create_position(row, col, fov)
                pos.create_image(
                    "0",
                    np.random.rand(2, len(channel_names), *zyx_shape).astype(
                        np.float32
                    ),
                )


@fixture(scope="session")
def preprocessed_hcs_dataset(tmp_path_factory: TempPathFactory) -> Path:
    """Provides a preprocessed HCS OME-Zarr dataset."""
    dataset_path = tmp_path_factory.mktemp("preprocessed.zarr")
    _build_hcs(dataset_path, channel_names, (12, 256, 256))
    # U[0, 1)
    expected = {"mean": 0.5, "std": 1 / np.sqrt(12), "median": 0.5, "iqr": 0.5}
    norm_meta = {channel: {"dataset_statistics": expected} for channel in channel_names}
    with open_ome_zarr(dataset_path, mode="r+") as dataset:
        dataset.zattrs["normalization"] = norm_meta
        for _, fov in dataset.positions():
            fov.zattrs["normalization"] = norm_meta
    return dataset_path


@fixture(scope="function")
def small_hcs_dataset(tmp_path_factory: TempPathFactory) -> Path:
    """Provides a small, not preprocessed HCS OME-Zarr dataset."""
    dataset_path = tmp_path_factory.mktemp("small.zarr")
    _build_hcs(dataset_path, channel_names, (12, 64, 64))
    return dataset_path



from pathlib import Path

from iohub.ngff import open_ome_zarr
from lightning.pytorch import LightningDataModule
from monai.transforms import Compose, MapTransform
from torch.utils.data import DataLoader

from viscy.data.hcs import ChannelMap, SlidingWindowDataset
from viscy.data.typing import Sample


class CTMCv1ValidationDataset(SlidingWindowDataset):

    def __len__(self, subsample_rate: int = 30) -> int:
        # sample every 30th frame in the videos
        return super().__len__() // self.subsample_rate

    def __getitem__(self, index: int) -> Sample:
        index = index * self.subsample_rate
        return super().__getitem__(index)


class CTMCv1DataModule(LightningDataModule):
    """
    Autoregression data module for the CTMCv1 dataset.
    Training and validation datasets are stored in separate HCS OME-Zarr stores.

    :param str | Path train_data_path: Path to the training dataset
    :param str | Path val_data_path: Path to the validation dataset
    :param list[MapTransform] train_transforms: List of transforms for training
    :param list[MapTransform] val_transforms: List of transforms for validation
    :param int batch_size: Batch size, defaults to 16
    :param int num_workers: Number of workers, defaults to 8
    :param str channel_name: Name of the DIC channel, defaults to "DIC"
    """

    def __init__(
        self,
        train_data_path: str | Path,
        val_data_path: str | Path,
        train_transforms: list[MapTransform],
        val_transforms: list[MapTransform],
        batch_size: int = 16,
        num_workers: int = 8,
        channel_name: str = "DIC",
    ) -> None:
        super().__init__()
        self.train_data_path = train_data_path
        self.val_data_path = val_data_path
        self.train_transforms = train_transforms
        self.val_transforms = val_transforms
        self.channel_map = ChannelMap(source=[channel_name], target=[channel_name])
        self.batch_size = batch_size
        self.num_workers = num_workers

    def setup(self, stage: str) -> None:
        if stage != "fit":
            raise NotImplementedError("Only fit stage is supported")
        self._setup_fit()

    def _setup_fit(self) -> None:
        train_plate = open_ome_zarr(self.train_data_path)
        val_plate = open_ome_zarr(self.val_data_path)
        train_positions = [p for _, p in train_plate.positions()]
        val_positions = [p for _, p in val_plate.positions()]
        self.train_dataset = SlidingWindowDataset(
            train_positions,
            channels=self.channel_map,
            z_window_size=1,
            transform=Compose(self.train_transforms),
        )
        self.val_dataset = CTMCv1ValidationDataset(
            val_positions,
            channels=self.channel_map,
            z_window_size=1,
            transform=Compose(self.val_transforms),
        )

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=bool(self.num_workers),
            shuffle=True,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=bool(self.num_workers),
            shuffle=False,
        )




# %% 
# %% Imports and paths

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torchview
import torchvision
from iohub import open_ome_zarr
from lightning.pytorch import seed_everything
from lightning.pytorch.loggers import CSVLogger

# pytorch lightning wrapper for Tensorboard.
from tensorboard import notebook  # for viewing tensorboard in notebook
from torch.utils.tensorboard import SummaryWriter  # for logging to tensorboard

# HCSDataModule makes it easy to load data during training.
from viscy.data.hcs import HCSDataModule

# Trainer class and UNet.
from viscy.light.engine import VSUNet
from viscy.light.trainer import VSTrainer

seed_everything(42, workers=True)

# Paths to data and log directory
data_path = Path(
    Path("~/data/04_image_translation/HEK_nuclei_membrane_pyramid.zarr/")
).expanduser()

log_dir = Path("~/data/04_image_translation/logs/").expanduser()

# Create log directory if needed, and launch tensorboard
log_dir.mkdir(parents=True, exist_ok=True)

# fmt: off
%reload_ext tensorboard
%tensorboard --logdir {log_dir} --port 6007 --bind_all
# fmt: on

# %%  The entire training loop is contained in this cell.

GPU_ID = 0
BATCH_SIZE = 10
YX_PATCH_SIZE = (512, 512)


# Dictionary that specifies key parameters of the model.
phase2fluor_config = {
    "architecture": "2D",
    "num_filters": [24, 48, 96, 192, 384],
    "in_channels": 1,
    "out_channels": 2,
    "residual": True,
    "dropout": 0.1,  # dropout randomly turns off weights to avoid overfitting of the model to data.
    "task": "reg",  # reg = regression task.
}

phase2fluor_model = VSUNet(
    model_config=phase2fluor_config.copy(),
    batch_size=BATCH_SIZE,
    loss_function=torch.nn.functional.l1_loss,
    schedule="WarmupCosine",
    log_num_samples=10,  # Number of samples from each batch to log to tensorboard.
    example_input_yx_shape=YX_PATCH_SIZE,
)

# Reinitialize the data module.
phase2fluor_data = HCSDataModule(
    data_path,
    source_channel="Phase",
    target_channel=["Nuclei", "Membrane"],
    z_window_size=1,
    split_ratio=0.8,
    batch_size=BATCH_SIZE,
    num_workers=8,
    architecture="2D",
    yx_patch_size=YX_PATCH_SIZE,
    augmentations=None,
)
phase2fluor_data.setup("fit")


# Train for 3 epochs to see if you can log graph.
trainer = VSTrainer(accelerator="gpu", devices=[GPU_ID], max_epochs=3, default_root_dir=log_dir)

# trainer class takes the model and the data module as inputs.
trainer.fit(phase2fluor_model, datamodule=phase2fluor_data)

# %% Is exmple_input_array present?
print(f'{phase2fluor_model.example_input_array.shape},{phase2fluor_model.example_input_array.dtype}')
trainer.logger.log_graph(phase2fluor_model, phase2fluor_model.example_input_array)
# %%



import argparse
from traitlets.config import Config
import nbformat as nbf
from nbconvert.preprocessors import TagRemovePreprocessor, ClearOutputPreprocessor
from nbconvert.exporters import NotebookExporter


def get_arg_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument('input_file')
    parser.add_argument('output_file')

    return parser


def convert(input_file, output_file):
    c = Config()
    c.TagRemovePreprocessor.remove_cell_tags = ("solution",)
    c.TagRemovePreprocessor.enabled = True
    c.ClearOutputPreprocesser.enabled = True
    c.NotebookExporter.preprocessors = [
        "nbconvert.preprocessors.TagRemovePreprocessor",
        "nbconvert.preprocessors.ClearOutputPreprocessor"
    ]

    exporter = NotebookExporter(config=c)
    exporter.register_preprocessor(TagRemovePreprocessor(config=c), True)
    exporter.register_preprocessor(ClearOutputPreprocessor(), True)

    output = NotebookExporter(config=c).from_filename(input_file)
    with open(output_file, 'w') as f:
        f.write(output[0])


if __name__ == "__main__":
    parser = get_arg_parser()
    args = parser.parse_args()

    convert(args.input_file, args.output_file)
    print(f'Converted {args.input_file} to {args.output_file}')



from enum import Enum
from typing import Literal, Sequence

from lightning.pytorch import LightningDataModule
from lightning.pytorch.utilities.combined_loader import CombinedLoader
from torch.utils.data import ConcatDataset, DataLoader

from viscy.data.hcs import _collate_samples


class CombineMode(Enum):
    MIN_SIZE = "min_size"
    MAX_SIZE_CYCLE = "max_size_cycle"
    MAX_SIZE = "max_size"
    SEQUENTIAL = "sequential"


class CombinedDataModule(LightningDataModule):
    """Wrapper for combining multiple data modules.
    For supported modes, see ``lightning.pytorch.utilities.combined_loader``.

    :param Sequence[LightningDataModule] data_modules: data modules to combine
    :param str train_mode: mode in training stage, defaults to "max_size_cycle"
    :param str val_mode: mode in validation stage, defaults to "sequential"
    :param str test_mode: mode in testing stage, defaults to "sequential"
    :param str predict_mode: mode in prediction stage, defaults to "sequential"
    """

    def __init__(
        self,
        data_modules: Sequence[LightningDataModule],
        train_mode: CombineMode = CombineMode.MAX_SIZE_CYCLE,
        val_mode: CombineMode = CombineMode.SEQUENTIAL,
        test_mode: CombineMode = CombineMode.SEQUENTIAL,
        predict_mode: CombineMode = CombineMode.SEQUENTIAL,
    ):
        super().__init__()
        self.data_modules = data_modules
        self.train_mode = CombineMode(train_mode).value
        self.val_mode = CombineMode(val_mode).value
        self.test_mode = CombineMode(test_mode).value
        self.predict_mode = CombineMode(predict_mode).value
        self.prepare_data_per_node = True

    def prepare_data(self):
        for dm in self.data_modules:
            dm.trainer = self.trainer
            dm.prepare_data()

    def setup(self, stage: Literal["fit", "validate", "test", "predict"]):
        for dm in self.data_modules:
            dm.setup(stage)

    def train_dataloader(self):
        return CombinedLoader(
            [dm.train_dataloader() for dm in self.data_modules], mode=self.train_mode
        )

    def val_dataloader(self):
        return CombinedLoader(
            [dm.val_dataloader() for dm in self.data_modules], mode=self.val_mode
        )

    def test_dataloader(self):
        return CombinedLoader(
            [dm.test_dataloader() for dm in self.data_modules], mode=self.test_mode
        )

    def predict_dataloader(self):
        return CombinedLoader(
            [dm.predict_dataloader() for dm in self.data_modules],
            mode=self.predict_mode,
        )


class ConcatDataModule(LightningDataModule):
    """
    Concatenate multiple data modules.
    The concatenated data module will have the same
    batch size and number of workers as the first data module.
    Each element will be sampled uniformly regardless of their original data module.
    :param Sequence[LightningDataModule] data_modules: data modules to concatenate
    """

    def __init__(self, data_modules: Sequence[LightningDataModule]):
        super().__init__()
        self.data_modules = data_modules
        self.num_workers = data_modules[0].num_workers
        self.batch_size = data_modules[0].batch_size
        for dm in data_modules:
            if dm.num_workers != self.num_workers:
                raise ValueError("Inconsistent number of workers")
            if dm.batch_size != self.batch_size:
                raise ValueError("Inconsistent batch size")
        self.prepare_data_per_node = True

    def prepare_data(self):
        for dm in self.data_modules:
            dm.trainer = self.trainer
            dm.prepare_data()

    def setup(self, stage: Literal["fit", "validate", "test", "predict"]):
        self.train_patches_per_stack = 0
        for dm in self.data_modules:
            dm.setup(stage)
            if patches := getattr(dm, "train_patches_per_stack", 0):
                if self.train_patches_per_stack == 0:
                    self.train_patches_per_stack = patches
                elif self.train_patches_per_stack != patches:
                    raise ValueError("Inconsistent patches per stack")
        if stage != "fit":
            raise NotImplementedError("Only fit stage is supported")
        self.train_dataset = ConcatDataset(
            [dm.train_dataset for dm in self.data_modules]
        )
        self.val_dataset = ConcatDataset([dm.val_dataset for dm in self.data_modules])

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size // self.train_patches_per_stack,
            num_workers=self.num_workers,
            shuffle=True,
            persistent_workers=bool(self.num_workers),
            collate_fn=_collate_samples,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
            persistent_workers=bool(self.num_workers),
        )



"""Redefine transforms from MONAI for jsonargparse."""

from typing import Sequence, Union

from monai.transforms import (
    MapTransform,
    RandAdjustContrastd,
    RandAffined,
    RandGaussianNoised,
    RandGaussianSmoothd,
    RandomizableTransform,
    RandScaleIntensityd,
    RandWeightedCropd,
)
from monai.transforms.transform import Randomizable
from numpy.random.mtrand import RandomState as RandomState
from typing_extensions import Iterable, Literal

from viscy.data.typing import Sample


class RandWeightedCropd(RandWeightedCropd):
    def __init__(
        self,
        keys: Union[Sequence[str], str],
        w_key: str,
        spatial_size: Sequence[int],
        num_samples: int = 1,
        **kwargs,
    ):
        super().__init__(
            keys=keys,
            w_key=w_key,
            spatial_size=spatial_size,
            num_samples=num_samples,
            **kwargs,
        )


class RandAffined(RandAffined):
    def __init__(
        self,
        keys: Union[Sequence[str], str],
        prob: float,
        rotate_range: Union[Sequence[float], float],
        shear_range: Union[Sequence[float], float],
        scale_range: Union[Sequence[float], float],
        **kwargs,
    ):
        super().__init__(
            keys=keys,
            prob=prob,
            rotate_range=rotate_range,
            shear_range=shear_range,
            scale_range=scale_range,
            **kwargs,
        )


class RandAdjustContrastd(RandAdjustContrastd):
    def __init__(
        self,
        keys: Union[Sequence[str], str],
        prob: float,
        gamma: Union[Sequence[float], float],
        **kwargs,
    ):
        super().__init__(
            keys=keys,
            prob=prob,
            gamma=gamma,
            **kwargs,
        )


class RandScaleIntensityd(RandScaleIntensityd):
    def __init__(
        self,
        keys: Union[Sequence[str], str],
        factors: Union[Sequence[float], float],
        prob: float,
        **kwargs,
    ):
        super().__init__(
            keys=keys,
            factors=factors,
            prob=prob,
            **kwargs,
        )


class RandGaussianNoised(RandGaussianNoised):
    def __init__(
        self,
        keys: Union[Sequence[str], str],
        prob: float,
        mean: Union[Sequence[float], float],
        std: Union[Sequence[float], float],
        **kwargs,
    ):
        super().__init__(
            keys=keys,
            prob=prob,
            mean=mean,
            std=std,
            **kwargs,
        )


class RandGaussianSmoothd(RandGaussianSmoothd):
    def __init__(
        self,
        keys: Union[Sequence[str], str],
        prob: float,
        sigma_x: Union[Sequence[float], float],
        sigma_y: Union[Sequence[float], float],
        sigma_z: Union[Sequence[float], float],
        **kwargs,
    ):
        super().__init__(
            keys=keys,
            prob=prob,
            sigma_x=sigma_x,
            sigma_y=sigma_y,
            sigma_z=sigma_z,
            **kwargs,
        )


class NormalizeSampled(MapTransform):
    """
    Normalize the sample
    :param Union[str, Iterable[str]] keys: keys to normalize
    :param str fov: fov path with respect to Plate
    :param str subtrahend: subtrahend for normalization, defaults to "mean"
    :param str divisor: divisor for normalization, defaults to "std"
    """

    def __init__(
        self,
        keys: Union[str, Iterable[str]],
        level: Literal["fov_statistics", "dataset_statistics"],
        subtrahend="mean",
        divisor="std",
    ) -> None:
        super().__init__(keys, allow_missing_keys=False)
        self.subtrahend = subtrahend
        self.divisor = divisor
        self.level = level

    # TODO: need to implement the case where the preprocessing already exists
    def __call__(self, sample: Sample) -> Sample:
        for key in self.keys:
            level_meta = sample["norm_meta"][key][self.level]
            subtrahend_val = level_meta[self.subtrahend]
            divisor_val = level_meta[self.divisor] + 1e-8  # avoid div by zero
            sample[key] = (sample[key] - subtrahend_val) / divisor_val
        return sample

    def _normalize():
        NotImplementedError("_normalization() not implemented")


class RandInvertIntensityd(MapTransform, RandomizableTransform):
    """
    Randomly invert the intensity of the image.
    """

    def __init__(
        self,
        keys: Union[str, Iterable[str]],
        prob: float = 0.1,
        allow_missing_keys: bool = False,
    ) -> None:
        MapTransform.__init__(self, keys, allow_missing_keys=allow_missing_keys)
        RandomizableTransform.__init__(self, prob)

    def __call__(self, sample: Sample) -> Sample:
        self.randomize(None)
        for key in self.keys:
            if key in sample:
                sample[key] = -sample[key]
        return sample

    def set_random_state(
        self, seed: int | None = None, state: RandomState | None = None
    ) -> Randomizable:
        super().set_random_state(seed, state)
        return self



import numpy as np
from torch.utils.tensorboard import SummaryWriter

import viscy.evaluation.evaluation_metrics as inference_metrics


class TorchEvaluator(object):
    """
    Handles all procedures involved with model evaluation.

    Params:
    :param dict torch_config: master config file
    """

    def __init__(self, torch_config, device=None) -> None:
        self.torch_config = torch_config

        self.zarr_dir = self.torch_config["zarr_dir"]
        self.network_config = self.torch_config["model"]
        self.training_config = self.torch_config["training"]
        self.dataset_config = self.torch_config["dataset"]
        self.inference_config = self.torch_config["inference"]
        self.preprocessing_config = self.torch_config["preprocessing"]

        self.inference_metrics = {}
        self.log_writer = SummaryWriter(log_dir=self.save_folder)

    def get_save_location(self):
        """
        Sets save location as specified in config files.
        """
        # TODO implement
        return
        # TODO Change the functionality of saving to put inference in the actual
        # train directory the model comes from. Not a big fan

        # model_dir = os.path.dirname(self.inference_config["model_dir"])
        # save_to_train_save_dir = self.inference_config["save_preds_to_model_dir"]

        # if save_to_train_save_dir:
        #     save_dir = model_dir
        # elif "custom_save_preds_dir" in self.inference_config:
        #     custom_save_dir = self.inference_config["custom_save_preds_dir"]
        #     save_dir = custom_save_dir
        # else:
        #     raise ValueError(
        #         "Must provide custom_save_preds_dir if save_preds_to"
        #         "_model_dir is False."
        #     )

        # now = aux_utils.get_timestamp()
        # self.save_folder = os.path.join(save_dir, f"inference_results_{now}")
        # if not os.path.exists(self.save_folder):
        #     os.makedirs(self.save_folder)

    def _collapse_metrics_dict(self, metrics_dict):
        """
        Collapses metrics dict in the form of
            {metric_name: {index: metric,...}}
        to the form
            {metric_name: np.ndarray[metric1, metrics2,...]}

        :param dict metrics_dict: dict of metrics in the first format

        :return dict collapsed_metrics_dict: dict of metrics in the second format
        """
        collapsed_metrics_dict = {}
        for metric_name in metrics_dict:
            val_dict = metrics_dict[metric_name]
            values = [val_dict[index] for index in val_dict]
            collapsed_metrics_dict[metric_name] = np.array(values)

        return collapsed_metrics_dict

    def _get_metrics(
        self,
        target,
        prediction,
        metrics_list,
        metrics_orientations,
        path="unspecified",
        window=None,
    ):
        """
        Gets metrics for this target_/prediction pair in all the specified orientations
        for all the specified metrics.

        :param np.ndarray target: 5d target array (on cpu)
        :param np.ndarray prediction: 5d prediction array (on cpu)
        :param list metrics_list: list of strings
            indicating the name of a desired metric,
            for options see inference.evaluation_metrics. MetricsEstimator docstring
        :param list metrics_orientations: list of strings
            indicating the orientation to compute,
            for options see inference.evaluation_metrics. MetricsEstimator docstring
        :param tuple window: spatial window of this target/prediction pair
            in the larger arrays they come from.

        :return dict prediction_metrics: dict mapping orientation -> pd.dataframe
            of metrics for that orientation
        """
        metrics_estimator = inference_metrics.MetricsEstimator(metrics_list)
        prediction_metrics = {}

        # transpose target and prediction to be in xyz format
        # NOTE: This expects target and pred to be in the format bczyx!
        target = np.transpose(target, (0, 1, -2, -1, -3))
        prediction = np.transpose(prediction, (0, 1, -2, -1, -3))

        zstart, zend = window[0][0], window[0][0] + window[1][0]  # end = start + length
        pred_name = f"slice_{zstart}-{zend}"

        if "xy" in metrics_orientations:
            metrics_estimator.estimate_xy_metrics(
                target=target,
                prediction=prediction,
                pred_name=pred_name,
            )
            metrics_xy = self._collapse_metrics_dict(
                metrics_estimator.get_metrics_xy().to_dict()
            )
            prediction_metrics["xy"] = metrics_xy

        if "xyz" in metrics_orientations:
            metrics_estimator.estimate_xyz_metrics(
                target=target,
                prediction=prediction,
                pred_name=pred_name,
            )
            metrics_xyz = self._collapse_metrics_dict(
                metrics_estimator.get_metrics_xyz().to_dict()
            )
            prediction_metrics["xyz"] = metrics_xyz

        if "xz" in metrics_orientations:
            metrics_estimator.estimate_xz_metrics(
                target=target,
                prediction=prediction,
                pred_name=pred_name,
            )
            metrics_xz = self._collapse_metrics_dict(
                metrics_estimator.get_metrics_xz().to_dict()
            )
            prediction_metrics["xz"] = metrics_xz

        if "yz" in metrics_orientations:
            metrics_estimator.estimate_yz_metrics(
                target=target,
                prediction=prediction,
                pred_name=pred_name,
            )
            metrics_yz = self._collapse_metrics_dict(
                metrics_estimator.get_metrics_yz().to_dict()
            )
            prediction_metrics["yz"] = metrics_yz

        # format metrics
        tag = path + f"_{window}"
        self.inference_metrics[tag] = prediction_metrics

        return prediction_metrics

    def record_metrics(self, sample_information):
        """
        Handles metric recording in tensorboard.

        Metrics are saved position by position.
        If multiple scalar metric values are stored for a
        particular metric in a particular position,
        they are plotted along the axis they are calculated on.

        :param list sample_information: list of tuples containing information about
            each sample in the form
            (position_group, position_path, normalization_meta, window)
        """
        for info_tuple in sample_information:
            _, position_path, normalization_meta, window = info_tuple
            position = position_path.split("/")[-1]
            sample_metrics = self.inference_metrics[position_path + f"_{window}"]

            for orientation in sample_metrics:
                scalar_dict = sample_metrics[orientation]
                pred_name = scalar_dict.pop("pred_name")[0]

                # generate a unique plot & tag for each orientation
                main_tag = f"{position}/{orientation}_{pred_name}"

                # Need to plot a line if metrics calculated along an axis
                if scalar_dict[list(scalar_dict.keys())[0]].shape[0] == 1:
                    self.writer.add_scalars(
                        main_tag=main_tag,
                        tag_scalar_dict=scalar_dict,
                    )
                else:
                    axis_length = scalar_dict[list(scalar_dict.keys())[0]].shape[0]
                    for i in range(axis_length):
                        scalar_dict_i = {}
                        for key in scalar_dict.keys():
                            scalar_dict_i[key] = scalar_dict[key][i]
                        self.writer.add_scalars(
                            main_tag=main_tag,
                            tag_scalar_dict=scalar_dict_i,
                            global_step=i,
                        )



# %% [markdown]
"""
# Image translation
---

Written by Ziwen Liu and Shalin Mehta, CZ Biohub San Francisco.

In this exercise, we will solve an image translation task to predict fluorescence images of nuclei and membrane markers from quantitative phase images of cells. In other words, we will _virtually stain_ the nuclei and membrane visible in the phase image. 

Here, the source domain is label-free microscopy (material density) and the target domain is fluorescence microscopy (fluorophore density). The goal is to learn a mapping from the source domain to the target domain. We will use a deep convolutional neural network (CNN), specifically, a U-Net model with residual connections to learn the mapping. The preprocessing, training, prediction, evaluation, and deployment steps are unified in a computer vision pipeline for single-cell analysis that we call [VisCy](https://github.com/mehta-lab/VisCy).

VisCy evolved from our previous work on virtual staining of cellular components from their density and anisotropy.
![](https://iiif.elifesciences.org/lax/55502%2Felife-55502-fig1-v2.tif/full/1500,/0/default.jpg)

[Guo et al. (2020) Revealing architectural order with quantitative label-free imaging and deep learning
. eLife](https://elifesciences.org/articles/55502).

VisCy exploits recent advances in the data and metadata formats ([OME-zarr](https://www.nature.com/articles/s41592-021-01326-w)) and DL frameworks, [PyTorch Lightning](https://lightning.ai/) and [MONAI](https://monai.io/). 
"""

# %% [markdown]
"""
Today, we will train a 2D image translation model using a 2D U-Net with residual connections. We will use a dataset of 301 fields of view (FOVs) of Human Embryonic Kidney (HEK) cells, each FOV has 3 channels (phase, membrane, and nuclei). The cells were labeled with CRISPR editing. Intrestingly, not all cells during this experiment were labeled due to the stochastic nature of CRISPR editing. In such situations, virtual staining rescues missing labels.
![HEK](https://github.com/mehta-lab/VisCy/blob/dlmbl2023/docs/figures/phase_to_nuclei_membrane.svg?raw=true)
"""
# %% [markdown]
"""
<div class="alert alert-warning">
The exercise is organized in 3 parts.

* **Part 1** - Explore the data using tensorboard. Launch the training before lunch.
* Lunch break - The model will continue training during lunch.
* **Part 2** - Evaluate the training with tensorboard. Train another model.
* **Part 3** - Tune the models to improve performance.
</div>
"""
# %% [markdown]
"""
📖 As you work through parts 2 and 3, please share the layouts of your models (output of torchview) and their performance with everyone via [this google doc](https://docs.google.com/document/d/1hZWSVRvt9KJEdYu7ib-vFBqAVQRYL8cWaP_vFznu7D8/edit#heading=h.n5u485pmzv2z) 📖.


Our guesstimate is that each of the three parts will take ~1.5 hours. A reasonable 2D UNet can be trained in ~20 min on a typical AWS node. 
We will discuss your observations on google doc after checkpoints 2 and 3.

The focus of the exercise is on understanding information content of the data, how to train and evaluate 2D image translation model, and explore some hyperparameters of the model. If you complete this exercise and have time to spare, try the bonus exercise on 3D image translation.
"""
# %% [markdown]
"""
<div class="alert alert-danger">
Set your python kernel to <span style="color:black;">04_image_translation</span>
</div>
"""
# %% <a [markdown] id='1_phase2fluor'></a>
"""
# Part 1: Log training data to tensorboard, start training a model.
---------

Learning goals:

- Load the OME-zarr dataset and examine the channels.
- Configure and understand the data loader.
- Log some patches to tensorboard.
- Initialize a 2D U-Net model for virtual staining
- Start training the model to predict nuclei and membrane from phase.
"""

# %% Imports and paths
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torchview
import torchvision
from iohub import open_ome_zarr
from lightning.pytorch import seed_everything
from lightning.pytorch.loggers import CSVLogger, TensorBoardLogger
from skimage import metrics  # for metrics.

# %% Imports and paths
# pytorch lightning wrapper for Tensorboard.
from torch.utils.tensorboard import SummaryWriter  # for logging to tensorboard

# HCSDataModule makes it easy to load data during training.
from viscy.data.hcs import HCSDataModule

# training augmentations
from viscy.transforms import (
    RandAdjustContrastd,
    RandAffined,
    RandGaussianNoised,
    RandGaussianSmoothd,
    RandScaleIntensityd,
    RandWeightedCropd,
)

# Trainer class and UNet.
from viscy.light.engine import VSUNet
from viscy.light.trainer import VSTrainer

seed_everything(42, workers=True)

# Paths to data and log directory
data_path = Path(
    Path("~/data/04_image_translation/HEK_nuclei_membrane_pyramid.zarr/")
).expanduser()

log_dir = Path("~/data/04_image_translation/logs/").expanduser()

# Create log directory if needed, and launch tensorboard
log_dir.mkdir(parents=True, exist_ok=True)

# %% [markdown] tags=[]
"""
The next cell starts tensorboard within the notebook.

<div class="alert alert-danger">
If you launched jupyter lab from ssh terminal, add <code>--host &lt;your-server-name&gt;</code> to the tensorboard command below. <code>&lt;your-server-name&gt;</code> is the address of your compute node that ends in amazonaws.com.

You can also launch tensorboard in an independent tab (instead of in the notebook) by changing the `%` to `!`
</div>
"""

# %% Imports and paths tags=[]
%reload_ext tensorboard
%tensorboard --logdir {log_dir}

# %% [markdown]
"""
## Load Dataset.

There should be 301 FOVs in the dataset (12 GB compressed).

Each FOV consists of 3 channels of 2048x2048 images,
saved in the <a href="https://ngff.openmicroscopy.org/latest/#hcs-layout">
High-Content Screening (HCS) layout</a>
specified by the Open Microscopy Environment Next Generation File Format
(OME-NGFF).

The layout on the disk is: row/col/field/pyramid_level/timepoint/channel/z/y/x.
Notice that labelling of nuclei channel is not complete - some cells are not expressing the fluorescent protein.
"""

# %%
dataset = open_ome_zarr(data_path)

print(f"Number of positions: {len(list(dataset.positions()))}")

# Use the field and pyramid_level below to visualize data.
row = 0
col = 0
field = 23  # TODO: Change this to explore data.

# This dataset contains images at 3 resolutions.
# '0' is the highest resolution
# '1' is down-scaled 2x2,
# '2' is down-scaled 4x4.
# Such datasets are called image pyramids.
pyaramid_level = 0

# `channel_names` is the metadata that is stored with data according to the OME-NGFF spec.
n_channels = len(dataset.channel_names)

image = dataset[f"{row}/{col}/{field}/{pyaramid_level}"].numpy()
print(f"data shape: {image.shape}, FOV: {field}, pyramid level: {pyaramid_level}")

figure, axes = plt.subplots(1, n_channels, figsize=(9, 3))

for i in range(n_channels):
    for i in range(n_channels):
        channel_image = image[0, i, 0]
        # Adjust contrast to 0.5th and 99.5th percentile of pixel values.
        p_low, p_high = np.percentile(channel_image, (0.5, 99.5))
        channel_image = np.clip(channel_image, p_low, p_high)
        axes[i].imshow(channel_image, cmap="gray")
        axes[i].axis("off")
        axes[i].set_title(dataset.channel_names[i])
plt.tight_layout()

# %% [markdown]
# <div class="alert alert-info">
#
# ### Task 1.1
#
# Look at a couple different fields of view by changing the value in the cell above. See if you notice any missing or inconsistent staining.
# </div>

# %% [markdown]
"""
## Explore the effects of augmentation on batch.

VisCy builds on top of PyTorch Lightning. PyTorch Lightning is a thin wrapper around PyTorch that allows rapid experimentation. It provides a [DataModule](https://lightning.ai/docs/pytorch/stable/data/datamodule.html) to handle loading and processing of data during training. VisCy provides a child class, `HCSDataModule` to make it intuitve to access data stored in the HCS layout.
  
The dataloader in `HCSDataModule` returns a batch of samples. A `batch` is a list of dictionaries. The length of the list is equal to the batch size. Each dictionary consists of following key-value pairs.
- `source`: the input image, a tensor of size 1*1*Y*X
- `target`: the target image, a tensor of size 2*1*Y*X
- `index` : the tuple of (location of field in HCS layout, time, and z-slice) of the sample.
"""

# %% [markdown]
# <div class="alert alert-info">
#
# ### Task 1.2
#
# Setup the data loader and log several batches to tensorboard.
#
# Based on the tensorboard images, what are the two channels in the target image?
#
# Note: If tensorboard is not showing images, try refreshing and using the "Images" tab.
# </div>

# %%
# Define a function to write a batch to tensorboard log.


def log_batch_tensorboard(batch, batchno, writer, card_name):
    """
    Logs a batch of images to TensorBoard.

    Args:
        batch (dict): A dictionary containing the batch of images to be logged.
        writer (SummaryWriter): A TensorBoard SummaryWriter object.
        card_name (str): The name of the card to be displayed in TensorBoard.

    Returns:
        None
    """
    batch_phase = batch["source"][:, :, 0, :, :]  # batch_size x z_size x Y x X tensor.
    batch_membrane = batch["target"][:, 1, 0, :, :].unsqueeze(
        1
    )  # batch_size x 1 x Y x X tensor.
    batch_nuclei = batch["target"][:, 0, 0, :, :].unsqueeze(
        1
    )  # batch_size x 1 x Y x X tensor.

    p1, p99 = np.percentile(batch_membrane, (0.1, 99.9))
    batch_membrane = np.clip((batch_membrane - p1) / (p99 - p1), 0, 1)

    p1, p99 = np.percentile(batch_nuclei, (0.1, 99.9))
    batch_nuclei = np.clip((batch_nuclei - p1) / (p99 - p1), 0, 1)

    p1, p99 = np.percentile(batch_phase, (0.1, 99.9))
    batch_phase = np.clip((batch_phase - p1) / (p99 - p1), 0, 1)

    [N, C, H, W] = batch_phase.shape
    interleaved_images = torch.zeros((3 * N, C, H, W), dtype=batch_phase.dtype)
    interleaved_images[0::3, :] = batch_phase
    interleaved_images[1::3, :] = batch_nuclei
    interleaved_images[2::3, :] = batch_membrane

    grid = torchvision.utils.make_grid(interleaved_images, nrow=3)

    # add the grid to tensorboard
    writer.add_image(card_name, grid, batchno)


# %%
# Define a function to visualize a batch on jupyter, in case tensorboard is finicky


def log_batch_jupyter(batch):
    """
    Logs a batch of images on jupyter using ipywidget.

    Args:
        batch (dict): A dictionary containing the batch of images to be logged.

    Returns:
        None
    """
    batch_phase = batch["source"][:, :, 0, :, :]  # batch_size x z_size x Y x X tensor.
    batch_size = batch_phase.shape[0]
    batch_membrane = batch["target"][:, 1, 0, :, :].unsqueeze(
        1
    )  # batch_size x 1 x Y x X tensor.
    batch_nuclei = batch["target"][:, 0, 0, :, :].unsqueeze(
        1
    )  # batch_size x 1 x Y x X tensor.

    p1, p99 = np.percentile(batch_membrane, (0.1, 99.9))
    batch_membrane = np.clip((batch_membrane - p1) / (p99 - p1), 0, 1)

    p1, p99 = np.percentile(batch_nuclei, (0.1, 99.9))
    batch_nuclei = np.clip((batch_nuclei - p1) / (p99 - p1), 0, 1)

    p1, p99 = np.percentile(batch_phase, (0.1, 99.9))
    batch_phase = np.clip((batch_phase - p1) / (p99 - p1), 0, 1)

    plt.figure()
    fig, axes = plt.subplots(batch_size, n_channels, figsize=(10, 10))
    [N, C, H, W] = batch_phase.shape
    for sample_id in range(batch_size):
        axes[sample_id, 0].imshow(batch_phase[sample_id, 0])
        axes[sample_id, 1].imshow(batch_nuclei[sample_id, 0])
        axes[sample_id, 2].imshow(batch_membrane[sample_id, 0])

        for i in range(n_channels):
            axes[sample_id, i].axis("off")
            axes[sample_id, i].set_title(dataset.channel_names[i])
    plt.tight_layout()
    plt.show()


# %%

# Initialize the data module.

BATCH_SIZE = 4
# 42 is a perfectly reasonable batch size. After all, it is the answer to the ultimate question of life, the universe and everything.
# More seriously, batch size does not have to be a power of 2.
# See: https://sebastianraschka.com/blog/2022/batch-size-2.html

data_module = HCSDataModule(
    data_path,
    source_channel="Phase",
    target_channel=["Membrane", "Nuclei"],
    z_window_size=1,
    split_ratio=0.8,
    batch_size=BATCH_SIZE,
    num_workers=8,
    architecture="2D",
    yx_patch_size=(512, 512),  # larger patch size makes it easy to see augmentations.
    augmentations=None,  # Turn off augmentation for now.
)
data_module.setup("fit")

print(
    f"FOVs in training set: {len(data_module.train_dataset)}, FOVs in validation set:{len(data_module.val_dataset)}"
)
train_dataloader = data_module.train_dataloader()

# Instantiate the tensorboard SummaryWriter, logs the first batch and then iterates through all the batches and logs them to tensorboard.

writer = SummaryWriter(log_dir=f"{log_dir}/view_batch")
# Draw a batch and write to tensorboard.
batch = next(iter(train_dataloader))
log_batch_tensorboard(batch, 0, writer, "augmentation/none")
writer.close()


# %% [markdown]
# Visualize directly on Jupyter ☄️, if your tensorboard is causing issues.

# %%
%matplotlib inline
log_batch_jupyter(batch)

# %% [markdown]
"""
## View augmentations using tensorboard.
"""
# %%
# Here we turn on data augmentation and rerun setup
augmentations = [
    RandWeightedCropd(
        keys=["Phase", "Membrane", "Nuclei"], w_key="Nuclei", spatial_size=[512, 512]
    ),
    RandAffined(
        keys=["Phase", "Membrane", "Nuclei"],
        prob=0.5,
        rotate_range=[3.14, 0.0, 0.0],
        shear_range=[0.0, 0.05, 0.05],
        scale_range=[0.0, 0.3, 0.3],
    ),
    RandAdjustContrastd(keys=["Phase"], prob=0.3, gamma=[0.5, 1.5]),
    RandScaleIntensityd(keys=["Phase"], prob=0.5, factors=0.5),
    RandGaussianNoised(keys=["Phase"], prob=0.5, std=1),
    RandGaussianSmoothd(
        keys=["Phase"], prob=0.5, sigma_x=[0.25, 1.5], sigma_y=[0.25, 1.5]
    ),
]

data_module.augmentations = augmentations
data_module.setup("fit")

# get the new data loader with augmentation turned on
augmented_train_dataloader = data_module.train_dataloader()

# Draw batches and write to tensorboard
writer = SummaryWriter(log_dir=f"{log_dir}/view_batch")
augmented_batch = next(iter(augmented_train_dataloader))
log_batch_tensorboard(augmented_batch, 0, writer, "augmentation/some")
writer.close()

# %% [markdown]
# Visualize directly on Jupyter ☄️

# %%
log_batch_jupyter(augmented_batch)

# %% [markdown]
# <div class="alert alert-info">
#
# ### Task 1.3
# Can you tell what augmentation were applied from looking at the augmented images in Tensorboard?
#
# Check your answer using the source code [here](https://github.com/mehta-lab/VisCy/blob/b89f778b34735553cf155904eef134c756708ff2/viscy/light/data.py#L529).
# </div>

# %% [markdown]
"""
## Train a 2D U-Net model to predict nuclei and membrane from phase.

### Construct a 2D U-Net
See ``viscy.unet.networks.Unet2D.Unet2d`` ([source code](https://github.com/mehta-lab/VisCy/blob/7c5e4c1d68e70163cf514d22c475da8ea7dc3a88/viscy/unet/networks/Unet2D.py#L7)) for configuration details.
"""
# %%
# Create a 2D UNet.
GPU_ID = 0
BATCH_SIZE = 10
YX_PATCH_SIZE = (512, 512)


# Dictionary that specifies key parameters of the model.
phase2fluor_config = {
    "architecture": "2D",
    "num_filters": [24, 48, 96, 192, 384],
    "in_channels": 1,
    "out_channels": 2,
    "residual": True,
    "dropout": 0.1,  # dropout randomly turns off weights to avoid overfitting of the model to data.
    "task": "reg",  # reg = regression task.
}

phase2fluor_model = VSUNet(
    model_config=phase2fluor_config.copy(),
    batch_size=BATCH_SIZE,
    loss_function=torch.nn.functional.l1_loss,
    schedule="WarmupCosine",
    log_num_samples=5,  # Number of samples from each batch to log to tensorboard.
    example_input_yx_shape=YX_PATCH_SIZE,
)


# %% [markdown]
"""
### Instantiate data module and trainer, test that we are setup to launch training.
"""
# %%
# Setup the data module.
phase2fluor_data = HCSDataModule(
    data_path,
    source_channel="Phase",
    target_channel=["Membrane", "Nuclei"],
    z_window_size=1,
    split_ratio=0.8,
    batch_size=BATCH_SIZE,
    num_workers=8,
    architecture="2D",
    yx_patch_size=YX_PATCH_SIZE,
    augmentations=augmentations,
)
phase2fluor_data.setup("fit")
# fast_dev_run runs a single batch of data through the model to check for errors.
trainer = VSTrainer(accelerator="gpu", devices=[GPU_ID], fast_dev_run=True)

# trainer class takes the model and the data module as inputs.
trainer.fit(phase2fluor_model, datamodule=phase2fluor_data)


# %% [markdown]
# ## View model graph.
#
# PyTorch uses dynamic graphs under the hood. The graphs are constructed on the fly. This is in contrast to TensorFlow, where the graph is constructed before the training loop and remains static. In other words, the graph of the network can change with every forward pass. Therefore, we need to supply an input tensor to construct the graph. The input tensor can be a random tensor of the correct shape and type. We can also supply a real image from the dataset. The latter is more useful for debugging.

# %% [markdown]
# <div class="alert alert-info">
#
# ### Task 1.4
# Run the next cell to generate a graph representation of the model architecture. Can you recognize the UNet structure and skip connections in this graph visualization?
# </div>

# %%
# visualize graph of phase2fluor model as image.
model_graph_phase2fluor = torchview.draw_graph(
    phase2fluor_model,
    phase2fluor_data.train_dataset[0]["source"],
    depth=2,  # adjust depth to zoom in.
    device="cpu",
)
# Print the image of the model.
model_graph_phase2fluor.visual_graph

# %% [markdown]
"""
<div class="alert alert-info">

### Task 1.5
Start training by running the following cell. Check the new logs on the tensorboard.
</div>
"""


# %%

GPU_ID = 0
n_samples = len(phase2fluor_data.train_dataset)
steps_per_epoch = n_samples // BATCH_SIZE  # steps per epoch.
n_epochs = 50  # Set this to 50 or the number of epochs you want to train for.

trainer = VSTrainer(
    accelerator="gpu",
    devices=[GPU_ID],
    max_epochs=n_epochs,
    log_every_n_steps=steps_per_epoch // 2,
    # log losses and image samples 2 times per epoch.
    logger=TensorBoardLogger(
        save_dir=log_dir,
        # lightning trainer transparently saves logs and model checkpoints in this directory.
        name="phase2fluor",
        log_graph=True,
    ),
)
# Launch training and check that loss and images are being logged on tensorboard.
trainer.fit(phase2fluor_model, datamodule=phase2fluor_data)

# %% [markdown]
"""
<div class="alert alert-success">

## Checkpoint 1

Now the training has started,
we can come back after a while and evaluate the performance!
</div>
"""

# %% <a [markdown] id='1_fluor2phase'></a>
"""
# Part 2: Assess previous model, train fluorescence to phase contrast translation model.
--------------------------------------------------
"""

# %% [markdown]
"""
We now look at some metrics of performance of previous model. We typically evaluate the model performance on a held out test data. We will use the following metrics to evaluate the accuracy of regression of the model:
- [Person Correlation](https://en.wikipedia.org/wiki/Pearson_correlation_coefficient).
- [Structural similarity](https://en.wikipedia.org/wiki/Structural_similarity) (SSIM).

You should also look at the validation samples on tensorboard (hint: the experimental data in nuclei channel is imperfect.)
"""

# %% [markdown]
"""
<div class="alert alert-info">

### Task 2.1 Define metrics

For each of the above metrics, write a brief definition of what they are and what they mean for this image translation task.

</div>
"""

# %% [markdown]
# ```
# #######################
# ##### Todo ############
# #######################
# ```
#
# - Pearson Correlation:
#
# - Structural similarity:

# %% Compute metrics directly and plot here.
test_data_path = Path(
    "~/data/04_image_translation/HEK_nuclei_membrane_test.zarr"
).expanduser()

test_data = HCSDataModule(
    test_data_path,
    source_channel="Phase",
    target_channel=["Membrane", "Nuclei"],
    z_window_size=1,
    batch_size=1,
    num_workers=8,
    architecture="2D",
)
test_data.setup("test")

test_metrics = pd.DataFrame(
    columns=["pearson_nuc", "SSIM_nuc", "pearson_mem", "SSIM_mem"]
)


def min_max_scale(input):
    return (input - np.min(input)) / (np.max(input) - np.min(input))


# %% Compute metrics directly and plot here.
for i, sample in enumerate(test_data.test_dataloader()):
    phase_image = sample["source"]
    with torch.inference_mode():  # turn off gradient computation.
        predicted_image = phase2fluor_model(phase_image)

    target_image = (
        sample["target"].cpu().numpy().squeeze(0)
    )  # Squeezing batch dimension.
    predicted_image = predicted_image.cpu().numpy().squeeze(0)
    phase_image = phase_image.cpu().numpy().squeeze(0)
    target_mem = min_max_scale(target_image[1, 0, :, :])
    target_nuc = min_max_scale(target_image[0, 0, :, :])
    # slicing channel dimension, squeezing z-dimension.
    predicted_mem = min_max_scale(predicted_image[1, :, :, :].squeeze(0))
    predicted_nuc = min_max_scale(predicted_image[0, :, :, :].squeeze(0))

    # Compute SSIM and pearson correlation.
    ssim_nuc = metrics.structural_similarity(target_nuc, predicted_nuc, data_range=1)
    ssim_mem = metrics.structural_similarity(target_mem, predicted_mem, data_range=1)
    pearson_nuc = np.corrcoef(target_nuc.flatten(), predicted_nuc.flatten())[0, 1]
    pearson_mem = np.corrcoef(target_mem.flatten(), predicted_mem.flatten())[0, 1]

    test_metrics.loc[i] = {
        "pearson_nuc": pearson_nuc,
        "SSIM_nuc": ssim_nuc,
        "pearson_mem": pearson_mem,
        "SSIM_mem": ssim_mem,
    }

test_metrics.boxplot(
    column=["pearson_nuc", "SSIM_nuc", "pearson_mem", "SSIM_mem"],
    rot=30,
)


# %% [markdown] tags=[]
"""
<div class="alert alert-info">

### Task 2.2 Train fluorescence to phase contrast translation model

Instantiate a data module, model, and trainer for fluorescence to phase contrast translation. Copy over the code from previous cells and update the parameters. Give the variables and paths a different name/suffix (fluor2phase) to avoid overwriting objects used to train phase2fluor models.
</div>
"""
# %% tags=[]
##########################
######## TODO ########
##########################

fluor2phase_data = HCSDataModule(
    # Your code here (copy from above and modify as needed)
)
fluor2phase_data.setup("fit")

# Dictionary that specifies key parameters of the model.
fluor2phase_config = {
    # Your config here
}

fluor2phase_model = VSUNet(
    # Your code here (copy from above and modify as needed)
)

trainer = VSTrainer(
    # Your code here (copy from above and modify as needed)
)
trainer.fit(fluor2phase_model, datamodule=fluor2phase_data)


# Visualize the graph of fluor2phase model as image.
model_graph_fluor2phase = torchview.draw_graph(
    fluor2phase_model,
    fluor2phase_data.train_dataset[0]["source"],
    depth=2,  # adjust depth to zoom in.
    device="cpu",
)
model_graph_fluor2phase.visual_graph

# %% tags=["solution"]

##########################
######## Solution ########
##########################

# The entire training loop is contained in this cell.

fluor2phase_data = HCSDataModule(
    data_path,
    source_channel="Membrane",
    target_channel="Phase",
    z_window_size=1,
    split_ratio=0.8,
    batch_size=BATCH_SIZE,
    num_workers=8,
    architecture="2D",
    yx_patch_size=YX_PATCH_SIZE,
    augmentations=augmentations,
)
fluor2phase_data.setup("fit")

# Dictionary that specifies key parameters of the model.
fluor2phase_config = {
    "architecture": "2D",
    "in_channels": 1,
    "out_channels": 1,
    "residual": True,
    "dropout": 0.1,  # dropout randomly turns off weights to avoid overfitting of the model to data.
    "task": "reg",  # reg = regression task.
    "num_filters": [24, 48, 96, 192, 384],
}

fluor2phase_model = VSUNet(
    model_config=fluor2phase_config.copy(),
    batch_size=BATCH_SIZE,
    loss_function=torch.nn.functional.mse_loss,
    schedule="WarmupCosine",
    log_num_samples=5,
    example_input_yx_shape=YX_PATCH_SIZE,
)


trainer = VSTrainer(
    accelerator="gpu",
    devices=[GPU_ID],
    max_epochs=n_epochs,
    log_every_n_steps=steps_per_epoch // 2,
    logger=TensorBoardLogger(
        save_dir=log_dir,
        # lightning trainer transparently saves logs and model checkpoints in this directory.
        name="fluor2phase",
        log_graph=True,
    ),
)
trainer.fit(fluor2phase_model, datamodule=fluor2phase_data)


# Visualize the graph of fluor2phase model as image.
model_graph_fluor2phase = torchview.draw_graph(
    fluor2phase_model,
    fluor2phase_data.train_dataset[0]["source"],
    depth=2,  # adjust depth to zoom in.
    device="cpu",
)
model_graph_fluor2phase.visual_graph

# %% [markdown] tags=[]
"""
<div class="alert alert-info">

### Task 2.3

While your model is training, let's think about the following questions:
- What is the information content of each channel in the dataset?
- How would you use image translation models?
- What can you try to improve the performance of each model?
</div>
"""
# %%
test_data_path = Path(
    "~/data/04_image_translation/HEK_nuclei_membrane_test.zarr"
).expanduser()

test_data = HCSDataModule(
    test_data_path,
    source_channel="Nuclei",  # or Membrane, depending on your choice of source
    target_channel="Phase",
    z_window_size=1,
    batch_size=1,
    num_workers=8,
    architecture="2D",
)
test_data.setup("test")

test_metrics = pd.DataFrame(columns=["pearson_phase", "SSIM_phase"])


def min_max_scale(input):
    return (input - np.min(input)) / (np.max(input) - np.min(input))


# %%
for i, sample in enumerate(test_data.test_dataloader()):
    source_image = sample["source"]
    with torch.inference_mode():  # turn off gradient computation.
        predicted_image = fluor2phase_model(source_image)

    target_image = (
        sample["target"].cpu().numpy().squeeze(0)
    )  # Squeezing batch dimension.
    predicted_image = predicted_image.cpu().numpy().squeeze(0)
    source_image = source_image.cpu().numpy().squeeze(0)
    target_phase = min_max_scale(target_image[0, 0, :, :])
    # slicing channel dimension, squeezing z-dimension.
    predicted_phase = min_max_scale(predicted_image[0, :, :, :].squeeze(0))

    # Compute SSIM and pearson correlation.
    ssim_phase = metrics.structural_similarity(
        target_phase, predicted_phase, data_range=1
    )
    pearson_phase = np.corrcoef(target_phase.flatten(), predicted_phase.flatten())[0, 1]

    test_metrics.loc[i] = {
        "pearson_phase": pearson_phase,
        "SSIM_phase": ssim_phase,
    }

test_metrics.boxplot(
    column=["pearson_phase", "SSIM_phase"],
    rot=30,
)

# %% [markdown] tags=[]
"""
<div class="alert alert-success">

## Checkpoint 2
When your model finishes training, please summarize hyperparameters and performance of your models in the [this google doc](https://docs.google.com/document/d/1hZWSVRvt9KJEdYu7ib-vFBqAVQRYL8cWaP_vFznu7D8/edit#heading=h.n5u485pmzv2z)

</div>
"""

# %% <a [markdown] id='3_tuning'></a> tags=[]
"""
# Part 3: Tune the models.
--------------------------------------------------

Learning goals: Understand how data, model capacity, and training parameters control the performance of the model. Your goal is to try to underfit or overfit the model.
"""


# %% [markdown] tags=[]
"""
<div class="alert alert-info">

### Task 3.1

- Choose a model you want to train (phase2fluor or fluor2phase).
- Set up a configuration that you think will improve the performance of the model
- Consider modifying the learning rate and see how it changes performance
- Use training loop illustrated in previous cells to train phase2fluor and fluor2phase models to prototype your own training loop.
- Add code to evaluate the model using Pearson Correlation and SSIM

As your model is training, please document hyperparameters, snapshots of predictions on validation set, and loss curves for your models in [this google doc](https://docs.google.com/document/d/1hZWSVRvt9KJEdYu7ib-vFBqAVQRYL8cWaP_vFznu7D8/edit#heading=h.n5u485pmzv2z)

</div>
"""
# %% tags=[]
##########################
######## TODO ########
##########################

tune_data = HCSDataModule(
    # Your code here (copy from above and modify as needed)
)
tune_data.setup("fit")

# Dictionary that specifies key parameters of the model.
tune_config = {
    # Your config here
}

tune_model = VSUNet(
    # Your code here (copy from above and modify as needed)
)

trainer = VSTrainer(
    # Your code here (copy from above and modify as needed)
)
trainer.fit(tune_model, datamodule=tune_data)


# Visualize the graph of fluor2phase model as image.
model_graph_tune = torchview.draw_graph(
    tune_model,
    tune_data.train_dataset[0]["source"],
    depth=2,  # adjust depth to zoom in.
    device="cpu",
)
model_graph_tune.visual_graph


# %% tags=["solution"]

##########################
######## Solution ########
##########################

phase2fluor_wider_config = {
    "architecture": "2D",
    # double the number of filters at each stage
    "num_filters": [48, 96, 192, 384, 768],
    "in_channels": 1,
    "out_channels": 2,
    "residual": True,
    "dropout": 0.1,
    "task": "reg",
}

phase2fluor_wider_model = VSUNet(
    model_config=phase2fluor_wider_config.copy(),
    batch_size=BATCH_SIZE,
    loss_function=torch.nn.functional.l1_loss,
    schedule="WarmupCosine",
    log_num_samples=5,
    example_input_yx_shape=YX_PATCH_SIZE,
)


trainer = VSTrainer(
    accelerator="gpu",
    devices=[GPU_ID],
    max_epochs=n_epochs,
    log_every_n_steps=steps_per_epoch,
    logger=TensorBoardLogger(
        save_dir=log_dir,
        name="phase2fluor",
        version="wider",
        log_graph=True,
    ),
    fast_dev_run=True,
)  # Set fast_dev_run to False to train the model.
trainer.fit(phase2fluor_wider_model, datamodule=phase2fluor_data)

# %% tags=["solution"]

##########################
######## Solution ########
##########################

phase2fluor_slow_model = VSUNet(
    model_config=phase2fluor_config.copy(),
    batch_size=BATCH_SIZE,
    loss_function=torch.nn.functional.l1_loss,
    # lower learning rate by 5 times
    lr=2e-4,
    schedule="WarmupCosine",
    log_num_samples=5,
    example_input_yx_shape=YX_PATCH_SIZE,
)

trainer = VSTrainer(
    accelerator="gpu",
    devices=[GPU_ID],
    max_epochs=n_epochs,
    log_every_n_steps=steps_per_epoch,
    logger=TensorBoardLogger(
        save_dir=log_dir,
        name="phase2fluor",
        version="low_lr",
        log_graph=True,
    ),
    fast_dev_run=True,
)
trainer.fit(phase2fluor_slow_model, datamodule=phase2fluor_data)


# %% [markdown] tags=[]
"""
<div class="alert alert-success">
    
## Checkpoint 3

Congratulations! You have trained several image translation models now!
Please document hyperparameters, snapshots of predictions on validation set, and loss curves for your models and add the final perforance in [this google doc](https://docs.google.com/document/d/1hZWSVRvt9KJEdYu7ib-vFBqAVQRYL8cWaP_vFznu7D8/edit#heading=h.n5u485pmzv2z). We'll discuss our combined results as a group.
</div>
"""



import logging
import os
from typing import Literal, Sequence, Union

import numpy as np
import torch
from imageio import imwrite
from lightning.pytorch import LightningModule
from matplotlib.pyplot import get_cmap
from monai.optimizers import WarmupCosineSchedule
from monai.transforms import DivisiblePad
from skimage.exposure import rescale_intensity
from torch import Tensor, nn
from torch.nn import functional as F
from torch.optim.lr_scheduler import ConstantLR
from torchmetrics.functional import (
    accuracy,
    cosine_similarity,
    dice,
    jaccard_index,
    mean_absolute_error,
    mean_squared_error,
    pearson_corrcoef,
    r2_score,
    structural_similarity_index_measure,
)

from viscy.data.hcs import Sample
from viscy.evaluation.evaluation_metrics import mean_average_precision, ms_ssim_25d
from viscy.unet.networks.fcmae import FullyConvolutionalMAE
from viscy.unet.networks.Unet2D import Unet2d
from viscy.unet.networks.Unet25D import Unet25d
from viscy.unet.networks.unext2 import UNeXt2

try:
    from cellpose.models import CellposeModel
except ImportError:
    CellposeModel = None


_UNET_ARCHITECTURE = {
    "2D": Unet2d,
    "UNeXt2": UNeXt2,
    "2.5D": Unet25d,
    "fcmae": FullyConvolutionalMAE,
}


class MixedLoss(nn.Module):
    """Mixed reconstruction loss.
    Adapted from Zhao et al, https://arxiv.org/pdf/1511.08861.pdf
    Reduces to simple distances if only one weight is non-zero.

    :param float l1_alpha: L1 loss weight, defaults to 0.5
    :param float l2_alpha: L2 loss weight, defaults to 0.0
    :param float ms_dssim_alpha: MS-DSSIM weight, defaults to 0.5
    """

    def __init__(
        self, l1_alpha: float = 0.5, l2_alpha: float = 0.0, ms_dssim_alpha: float = 0.5
    ):
        super().__init__()
        if not any([l1_alpha, l2_alpha, ms_dssim_alpha]):
            raise ValueError("Loss term weights cannot be all zero!")
        self.l1_alpha = l1_alpha
        self.l2_alpha = l2_alpha
        self.ms_dssim_alpha = ms_dssim_alpha

    @torch.cuda.amp.custom_fwd(cast_inputs=torch.float32)
    def forward(self, preds, target):
        loss = 0
        if self.l1_alpha:
            # the gaussian in the reference is not used
            # because the SSIM here uses a uniform window
            loss += F.l1_loss(preds, target) * self.l1_alpha
        if self.l2_alpha:
            loss += F.mse_loss(preds, target) * self.l2_alpha
        if self.ms_dssim_alpha:
            ms_ssim = ms_ssim_25d(preds, target, clamp=True)
            # the 1/2 factor in the original DSSIM is not used
            # since the MS-SSIM here is stabilized with ReLU
            loss += (1 - ms_ssim) * self.ms_dssim_alpha
        return loss


class VSUNet(LightningModule):
    """Regression U-Net module for virtual staining.

    :param dict model_config: model config,
        defaults to :py:class:`viscy.unet.utils.model.ModelDefaults25D`
    :param Union[nn.Module, MixedLoss] loss_function:
        loss function in training/validation,
        if a dictionary, should specify weights of each term
        ('l1_alpha', 'l2_alpha', 'ssim_alpha')
        defaults to L2 (mean squared error)
    :param float lr: learning rate in training, defaults to 1e-3
    :param Literal['WarmupCosine', 'Constant'] schedule:
        learning rate scheduler, defaults to "Constant"
    :param str chkpt_path: path to the checkpoint to load weights, defaults to None
    :param int log_batches_per_epoch:
        number of batches to log each training/validation epoch,
        has to be smaller than steps per epoch, defaults to 8
    :param int log_samples_per_batch:
        number of samples to log each training/validation batch,
        has to be smaller than batch size, defaults to 1
    :param Sequence[int] example_input_yx_shape:
        XY shape of the example input for network graph tracing, defaults to (256, 256)
    :param str test_cellpose_model_path:
        path to the CellPose model for testing segmentation, defaults to None
    :param float test_cellpose_diameter:
        diameter parameter of the CellPose model for testing segmentation,
        defaults to None
    :param bool test_evaluate_cellpose:
        evaluate the performance of the CellPose model instead of the trained model
        in test stage, defaults to False
    """

    def __init__(
        self,
        architecture: Literal["2D", "UNeXt2", "2.5D", "3D", "fcmae"],
        model_config: dict = {},
        loss_function: Union[nn.Module, MixedLoss] = None,
        lr: float = 1e-3,
        schedule: Literal["WarmupCosine", "Constant"] = "Constant",
        freeze_encoder: bool = False,
        ckpt_path: str = None,
        log_batches_per_epoch: int = 8,
        log_samples_per_batch: int = 1,
        example_input_yx_shape: Sequence[int] = (256, 256),
        test_cellpose_model_path: str = None,
        test_cellpose_diameter: float = None,
        test_evaluate_cellpose: bool = False,
    ) -> None:
        super().__init__()
        net_class = _UNET_ARCHITECTURE.get(architecture)
        if not net_class:
            raise ValueError(
                f"Architecture {architecture} not in {_UNET_ARCHITECTURE.keys()}"
            )
        self.model = net_class(**model_config)
        # TODO: handle num_outputs in metrics
        # self.out_channels = self.model.terminal_block.out_filters
        self.loss_function = loss_function if loss_function else nn.MSELoss()
        self.lr = lr
        self.schedule = schedule
        self.log_batches_per_epoch = log_batches_per_epoch
        self.log_samples_per_batch = log_samples_per_batch
        self.training_step_outputs = []
        self.validation_losses = []
        self.validation_step_outputs = []
        # required to log the graph
        if architecture == "2D":
            example_depth = 1
        else:
            example_depth = model_config.get("in_stack_depth") or 5
        self.example_input_array = torch.rand(
            1,
            model_config.get("in_channels") or 1,
            example_depth,
            *example_input_yx_shape,
        )
        self.test_cellpose_model_path = test_cellpose_model_path
        self.test_cellpose_diameter = test_cellpose_diameter
        self.test_evaluate_cellpose = test_evaluate_cellpose
        self.freeze_encoder = freeze_encoder
        if ckpt_path is not None:
            self.load_state_dict(
                torch.load(ckpt_path)["state_dict"]
            )  # loading only weights

    def forward(self, x: Tensor) -> Tensor:
        return self.model(x)

    def training_step(self, batch: Sample | Sequence[Sample], batch_idx: int):
        losses = []
        batch_size = 0
        if not isinstance(batch, Sequence):
            batch = [batch]
        for b in batch:
            source = b["source"]
            target = b["target"]
            pred = self.forward(source)
            loss = self.loss_function(pred, target)
            losses.append(loss)
            batch_size += source.shape[0]
            if batch_idx < self.log_batches_per_epoch:
                self.training_step_outputs.extend(
                    self._detach_sample((source, target, pred))
                )
        loss_step = torch.stack(losses).mean()
        self.log(
            "loss/train",
            loss_step.to(self.device),
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
            batch_size=batch_size,
        )
        return loss_step

    def validation_step(self, batch: Sample, batch_idx: int, dataloader_idx: int = 0):
        source: Tensor = batch["source"]
        target: Tensor = batch["target"]
        pred = self.forward(source)
        loss = self.loss_function(pred, target)
        if dataloader_idx + 1 > len(self.validation_losses):
            self.validation_losses.append([])
        self.validation_losses[dataloader_idx].append(loss.detach())
        self.log(
            f"loss/val/{dataloader_idx}",
            loss.to(self.device),
            sync_dist=True,
            batch_size=source.shape[0],
        )
        if batch_idx < self.log_batches_per_epoch:
            self.validation_step_outputs.extend(
                self._detach_sample((source, target, pred))
            )

    def test_step(self, batch: Sample, batch_idx: int):
        source = batch["source"]
        target = batch["target"]
        center_index = target.shape[-3] // 2
        center_slice = slice(center_index, center_index + 1)
        target = target[:, 0, center_slice]
        if self.test_evaluate_cellpose:
            pred = target
        else:
            pred = self.forward(source)[:, 0, center_slice]
            # FIXME: Only works for batch size 1 and the first channel
            self._log_regression_metrics(pred, target)
        img_names, ts, zs = batch["index"]
        position = float(img_names[0].split("/")[-2])
        self.log_dict(
            {
                "position": position,
                "time": float(ts[0]),
                "slice": float(zs[0]),
            },
            on_step=True,
            on_epoch=False,
        )
        if "labels" in batch:
            pred_labels = self._cellpose_predict(
                pred, f"p{int(position)}_t{ts[0]}_z{zs[0]}"
            )
            self._log_segmentation_metrics(pred_labels, batch["labels"][0])
        else:
            self._log_segmentation_metrics(None, None)

    def _log_regression_metrics(self, pred: Tensor, target: Tensor):
        # paired image translation metrics
        self.log_dict(
            {
                # regression
                "test_metrics/MAE": mean_absolute_error(pred, target),
                "test_metrics/MSE": mean_squared_error(pred, target),
                "test_metrics/cosine": cosine_similarity(
                    pred, target, reduction="mean"
                ),
                "test_metrics/pearson": pearson_corrcoef(
                    pred.flatten() * 1e4, target.flatten() * 1e4
                ),
                "test_metrics/r2": r2_score(pred.flatten(), target.flatten()),
                # image perception
                "test_metrics/SSIM": structural_similarity_index_measure(
                    pred, target, gaussian_kernel=False, kernel_size=21
                ),
            },
            on_step=True,
            on_epoch=True,
        )

    def _cellpose_predict(self, pred: Tensor, name: str) -> torch.ShortTensor:
        pred_labels_np = self.cellpose_model.eval(
            pred.cpu().numpy(), channels=[0, 0], diameter=self.test_cellpose_diameter
        )[0].astype(np.int16)
        imwrite(os.path.join(self.logger.log_dir, f"{name}.png"), pred_labels_np)
        return torch.from_numpy(pred_labels_np).to(self.device)

    def _log_segmentation_metrics(
        self, pred_labels: torch.ShortTensor, target_labels: torch.ShortTensor
    ):
        compute = pred_labels is not None
        if compute:
            pred_binary = pred_labels > 0
            target_binary = target_labels > 0
            coco_metrics = mean_average_precision(pred_labels, target_labels)
            logging.debug(coco_metrics)
        self.log_dict(
            {
                # semantic segmentation
                "test_metrics/accuracy": (
                    accuracy(pred_binary, target_binary, task="binary")
                    if compute
                    else -1
                ),
                "test_metrics/dice": (
                    dice(pred_binary, target_binary) if compute else -1
                ),
                "test_metrics/jaccard": (
                    jaccard_index(pred_binary, target_binary, task="binary")
                    if compute
                    else -1
                ),
                "test_metrics/mAP": coco_metrics["map"] if compute else -1,
                "test_metrics/mAP_50": coco_metrics["map_50"] if compute else -1,
                "test_metrics/mAP_75": coco_metrics["map_75"] if compute else -1,
                "test_metrics/mAR_100": coco_metrics["mar_100"] if compute else -1,
            },
            on_step=True,
            on_epoch=False,
        )

    def predict_step(self, batch: Sample, batch_idx: int, dataloader_idx: int = 0):
        source = self._predict_pad(batch["source"])
        return self._predict_pad.inverse(self.forward(source))

    def on_train_epoch_end(self):
        self._log_samples("train_samples", self.training_step_outputs)
        self.training_step_outputs = []

    def on_validation_epoch_end(self):
        super().on_validation_epoch_end()
        self._log_samples("val_samples", self.validation_step_outputs)
        self.validation_step_outputs = []
        # average within each dataloader
        loss_means = [torch.tensor(losses).mean() for losses in self.validation_losses]
        self.log(
            "loss/validate",
            torch.tensor(loss_means).mean().to(self.device),
            sync_dist=True,
        )

    def on_test_start(self):
        """Load CellPose model for segmentation."""
        if CellposeModel is None:
            # raise ImportError(
            #     "CellPose not installed. "
            #     "Please install the metrics dependency with "
            #     '`pip install viscy".[metrics]"`'
            # )
            logging.warning(
                "CellPose not installed. "
                "Please install the metrics dependency with "
                '`pip install viscy"[metrics]"`'
            )

        if self.test_cellpose_model_path is not None:
            self.cellpose_model = CellposeModel(
                model_type=self.test_cellpose_model_path, device=self.device
            )

    def on_predict_start(self):
        """Pad the input shape to be divisible by the downsampling factor.
        The inverse of this transform crops the prediction to original shape.
        """
        down_factor = 2**self.model.num_blocks
        self._predict_pad = DivisiblePad((0, 0, down_factor, down_factor))

    def configure_optimizers(self):
        if self.freeze_encoder:
            self.model: FullyConvolutionalMAE
            self.model.encoder.requires_grad_(False)
        optimizer = torch.optim.AdamW(self.model.parameters(), lr=self.lr)
        if self.schedule == "WarmupCosine":
            scheduler = WarmupCosineSchedule(
                optimizer,
                warmup_steps=3,
                t_total=self.trainer.max_epochs,
                warmup_multiplier=1e-3,
            )
        elif self.schedule == "Constant":
            scheduler = ConstantLR(
                optimizer, factor=1, total_iters=self.trainer.max_epochs
            )
        return [optimizer], [scheduler]

    def _detach_sample(self, imgs: Sequence[Tensor]):
        num_samples = min(imgs[0].shape[0], self.log_samples_per_batch)
        return [
            [np.squeeze(img[i].detach().cpu().numpy().max(axis=1)) for img in imgs]
            for i in range(num_samples)
        ]

    def _log_samples(self, key: str, imgs: Sequence[Sequence[np.ndarray]]):
        images_grid = []
        for sample_images in imgs:
            images_row = []
            for i, image in enumerate(sample_images):
                cm_name = "gray" if i == 0 else "inferno"
                if image.ndim == 2:
                    image = image[np.newaxis]
                for channel in image:
                    channel = rescale_intensity(channel, out_range=(0, 1))
                    render = get_cmap(cm_name)(channel, bytes=True)[..., :3]
                    images_row.append(render)
            images_grid.append(np.concatenate(images_row, axis=1))
        grid = np.concatenate(images_grid, axis=0)
        self.logger.experiment.add_image(
            key, grid, self.current_epoch, dataformats="HWC"
        )


class FcmaeUNet(VSUNet):
    def __init__(self, fit_mask_ratio: float = 0.0, **kwargs):
        super().__init__(architecture="fcmae", **kwargs)
        self.fit_mask_ratio = fit_mask_ratio

    def forward(self, x: Tensor, mask_ratio: float = 0.0):
        return self.model(x, mask_ratio)

    def forward_fit(self, batch: Sample) -> tuple[Tensor]:
        source = batch["source"]
        target = batch["target"]
        pred, mask = self.forward(source, mask_ratio=self.fit_mask_ratio)
        loss = F.mse_loss(pred, target, reduction="none")
        loss = (loss.mean(2) * mask).sum() / mask.sum()
        return source, target, pred, mask, loss

    def training_step(self, batch: Sequence[Sample], batch_idx: int):
        losses = []
        batch_size = 0
        for b in batch:
            source, target, pred, mask, loss = self.forward_fit(b)
            losses.append(loss)
            batch_size += source.shape[0]
            if batch_idx < self.log_batches_per_epoch:
                self.training_step_outputs.extend(
                    self._detach_sample((source, target * mask.unsqueeze(2), pred))
                )
        loss_step = torch.stack(losses).mean()
        self.log(
            "loss/train",
            loss_step.to(self.device),
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
            batch_size=batch_size,
        )
        return loss_step

    def validation_step(self, batch: Sample, batch_idx: int, dataloader_idx: int = 0):
        source, target, pred, mask, loss = self.forward_fit(batch)
        if dataloader_idx + 1 > len(self.validation_losses):
            self.validation_losses.append([])
        self.validation_losses[dataloader_idx].append(loss.detach())
        self.log(
            f"loss/val/{dataloader_idx}",
            loss.to(self.device),
            sync_dist=True,
            batch_size=source.shape[0],
        )
        if batch_idx < self.log_batches_per_epoch:
            self.validation_step_outputs.extend(
                self._detach_sample((source, target * mask.unsqueeze(2), pred))
            )



import json
from pathlib import Path

import torch
from lightning.pytorch import LightningDataModule
from monai.transforms import Compose, MapTransform
from pycocotools.coco import COCO
from tifffile import imread
from torch.utils.data import DataLoader, Dataset
from torchvision.ops import box_convert

from viscy.data.typing import Sample


class LiveCellDataset(Dataset):
    """
    LiveCell dataset.

    :param list[Path] images: List of paths to single-page, single-channel TIFF files.
    :param MapTransform | Compose transform: Transform to apply to the dataset
    """

    def __init__(self, images: list[Path], transform: MapTransform | Compose) -> None:
        self.images = images
        self.transform = transform

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Sample:
        image = imread(self.images[idx])[None, None]
        image = torch.from_numpy(image).to(torch.float32)
        image = self.transform(image)
        return {"source": image, "target": image}


class LiveCellTestDataset(Dataset):
    """
    LiveCell dataset.

    :param list[Path] images: List of paths to single-page, single-channel TIFF files.
    :param MapTransform | Compose transform: Transform to apply to the dataset
    """

    def __init__(
        self,
        image_dir: Path,
        transform: MapTransform | Compose,
        annotations: Path,
        load_target: bool = False,
        load_labels: bool = False,
    ) -> None:
        self.image_dir = image_dir
        self.transform = transform
        self.coco = COCO(str(annotations))
        self.image_ids = list(self.coco.imgs.keys())
        self.load_target = load_target
        self.load_labels = load_labels

    def __len__(self) -> int:
        return len(self.image_ids)

    def __getitem__(self, idx: int) -> Sample:
        image_id = self.image_ids[idx]
        file_name = self.coco.imgs[image_id]["file_name"]
        image_path = self.image_dir / file_name
        image = imread(image_path)[None, None]
        image = torch.from_numpy(image).to(torch.float32)
        sample = Sample(source=image)
        if self.load_target:
            sample["target"] = image
        if self.load_labels:
            anns = self.coco.loadAnns(self.coco.getAnnIds(image_id)) or []
            boxes = [torch.tensor(ann["bbox"]).to(torch.float32) for ann in anns]
            masks = [
                torch.from_numpy(self.coco.annToMask(ann)).to(torch.bool)
                for ann in anns
            ]
            dets = {
                "boxes": box_convert(torch.stack(boxes), in_fmt="xywh", out_fmt="xyxy"),
                "labels": torch.zeros(len(anns)).to(torch.uint8),
                "masks": torch.stack(masks),
            }
            sample["detections"] = dets
            sample["file_name"] = file_name
        self.transform(sample)
        return sample


class LiveCellDataModule(LightningDataModule):
    def __init__(
        self,
        train_val_images: Path | None = None,
        test_images: Path | None = None,
        train_annotations: Path | None = None,
        val_annotations: Path | None = None,
        test_annotations: Path | None = None,
        train_transforms: list[MapTransform] = [],
        val_transforms: list[MapTransform] = [],
        test_transforms: list[MapTransform] = [],
        batch_size: int = 16,
        num_workers: int = 8,
    ) -> None:
        super().__init__()
        self.train_val_images = Path(train_val_images)
        if not self.train_val_images.is_dir():
            raise NotADirectoryError(str(train_val_images))
        self.test_images = Path(test_images)
        if not self.test_images.is_dir():
            raise NotADirectoryError(str(test_images))
        self.train_annotations = Path(train_annotations)
        if not self.train_annotations.is_file():
            raise FileNotFoundError(str(train_annotations))
        self.val_annotations = Path(val_annotations)
        if not self.val_annotations.is_file():
            raise FileNotFoundError(str(val_annotations))
        self.test_annotations = Path(test_annotations)
        if not self.test_annotations.is_file():
            raise FileNotFoundError(str(test_annotations))
        self.train_transforms = Compose(train_transforms)
        self.val_transforms = Compose(val_transforms)
        self.test_transforms = Compose(test_transforms)
        self.batch_size = batch_size
        self.num_workers = num_workers

    def setup(self, stage: str) -> None:
        if stage == "fit":
            self._setup_fit()
        elif stage == "test":
            self._setup_test()

    def _parse_image_names(self, annotations: Path) -> list[Path]:
        with open(annotations) as f:
            images = [f["file_name"] for f in json.load(f)["images"]]
        return sorted(images)

    def _setup_fit(self) -> None:
        train_images = self._parse_image_names(self.train_annotations)
        val_images = self._parse_image_names(self.val_annotations)
        self.train_dataset = LiveCellDataset(
            [self.train_val_images / f for f in train_images],
            transform=self.train_transforms,
        )
        self.val_dataset = LiveCellDataset(
            [self.train_val_images / f for f in val_images],
            transform=self.val_transforms,
        )

    def _setup_test(self) -> None:
        self.test_dataset = LiveCellTestDataset(
            self.test_images,
            transform=self.test_transforms,
            annotations=self.test_annotations,
            load_labels=True,
        )

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=bool(self.num_workers),
            shuffle=True,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=bool(self.num_workers),
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset, batch_size=self.batch_size, num_workers=self.num_workers
        )



# %% script to generate your ground truth directory for viscy prediction evaluation
# After inference, the predictions generated are stored as zarr store.
# Evaluation metrics can be computed by comparison of prediction to
# human proof read ground truth.

import argparse
import os

import imageio as iio
import iohub.ngff as ngff
import pandas as pd

import viscy.evaluation.evaluation_metrics as metrics
import viscy.utils.aux_utils as aux_utils

# %% read the below details from the config file


def parse_args():
    """
    Parse command line arguments
    In python namespaces are implemented as dictionaries

    :return: namespace containing the arguments passed.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--config",
        type=str,
        help="path to yaml configuration file",
    )
    args = parser.parse_args()
    return args


def main(config):
    """
    pick focus slice mask from pred_zarr from slice number stored on png mask name
    input pred mask & corrected ground truth mask to metrics computation
    store the metrics values as csv file to corresponding positions in list
    Info to be stored:
        1. position no,
        2. eval metrics values
    """

    torch_config = aux_utils.read_config(config)

    pred_dir = torch_config["evaluation_metrics"]["pred_dir"]
    metric_channel = torch_config["evaluation_metrics"]["metric_channel"]
    PosList = torch_config["evaluation_metrics"]["PosList"]
    z_list = torch_config["evaluation_metrics"]["z_list"]
    metrics_list = torch_config["evaluation_metrics"]["metrics"]
    ground_truth_chans = torch_config["data"]["target_channel"]
    ground_truth_subdir = "ground_truth"

    d_pod = [
        "OD_true_positives",
        "OD_false_positives",
        "OD_false_negatives",
        "OD_precision",
        "OD_recall",
        "OD_f1_score",
    ]

    metric_map = {
        "ssim": metrics.ssim_metric,
        "corr": metrics.corr_metric,
        "r2": metrics.r2_metric,
        "mse": metrics.mse_metric,
        "mae": metrics.mae_metric,
        "dice": metrics.dice_metric,
        "IoU": metrics.IOU_metric,
        "VI": metrics.VOI_metric,
        "POD": metrics.POD_metric,
    }

    path_split_head_tail = os.path.split(pred_dir)
    target_zarr_dir = path_split_head_tail[0]
    pred_plate = ngff.open_ome_zarr(
        store_path=os.path.join(target_zarr_dir, metric_channel + "_pred.zarr"),
        mode="r+",
    )
    chan_names = pred_plate.channel_names
    metric_chan_mask = metric_channel + "_cp_mask"
    ground_truth_dir = os.path.join(target_zarr_dir, ground_truth_subdir)

    col_val = metrics_list[:]
    if "POD" in col_val:
        col_val.remove("POD")
        for i in range(len(d_pod)):
            col_val.insert(i + metrics_list.index("POD"), d_pod[i])
    df_metrics = pd.DataFrame(columns=col_val, index=PosList)

    for position, pos_data in pred_plate.positions():
        pos = int(position.split("/")[-1])

        if pos in PosList:
            idx = PosList.index(pos)
            raw_data = pos_data.data
            pred_mask = raw_data[0, chan_names.index(metric_chan_mask)]

            z_slice_no = z_list[idx]
            gt_mask_save_name = (
                ground_truth_chans[0]
                + "_p"
                + str(format(pos, "03d"))
                + "_z"
                + str(z_slice_no)
                + "_cp_mask.png"
            )

            gt_mask = iio.imread(os.path.join(ground_truth_dir, gt_mask_save_name))

            pos_metric_list = []
            for metric_name in metrics_list:
                metric_fn = metric_map[metric_name]
                cur_metric_list = metric_fn(
                    gt_mask,
                    pred_mask[0],
                )
                pos_metric_list = pos_metric_list + cur_metric_list

            df_metrics.loc[pos] = pos_metric_list

    csv_filename = os.path.join(ground_truth_dir, "GT_metrics.csv")
    df_metrics.to_csv(csv_filename)


if __name__ == "__main__":
    args = parse_args()
    main(args.config)



import logging
import os
import sys
from datetime import datetime

import torch
from jsonargparse import lazy_instance
from lightning.pytorch import LightningModule
from lightning.pytorch.cli import LightningCLI
from lightning.pytorch.loggers import TensorBoardLogger

from viscy.data.hcs import HCSDataModule
from viscy.light.engine import VSUNet
from viscy.light.trainer import VSTrainer


class VSLightningCLI(LightningCLI):
    """Extending lightning CLI arguments and defualts."""

    @staticmethod
    def subcommands() -> dict[str, set[str]]:
        subcommands = LightningCLI.subcommands()
        subcommands["preprocess"] = {"model", "dataloaders", "datamodule"}
        subcommands["export"] = {"model", "dataloaders", "datamodule"}
        return subcommands

    def add_arguments_to_parser(self, parser):
        if "preprocess" not in sys.argv:
            parser.link_arguments("data.yx_patch_size", "model.example_input_yx_shape")
            parser.link_arguments("model.architecture", "data.architecture")
        parser.set_defaults(
            {
                "trainer.logger": lazy_instance(
                    TensorBoardLogger,
                    save_dir="",
                    version=datetime.now().strftime(r"%Y%m%d-%H%M%S"),
                    log_graph=True,
                )
            }
        )


def main():
    """Main Lightning CLI entry point."""
    log_level = os.getenv("VISCY_LOG_LEVEL", logging.INFO)
    logging.getLogger("lightning.pytorch").setLevel(log_level)
    torch.set_float32_matmul_precision("high")
    model_class = VSUNet
    datamodule_class = HCSDataModule
    seed = True
    if "preprocess" in sys.argv:
        seed = False
        model_class = LightningModule
        datamodule_class = None
    _ = VSLightningCLI(
        model_class=model_class,
        datamodule_class=datamodule_class,
        trainer_class=VSTrainer,
        seed_everything_default=seed,
    )


if __name__ == "__main__":
    main()
